
import re
from pwpeg import *

from random import randint

def _choose(lbracket, text, repeat, rbracket):
    l = [text] + repeat
    return l[randint(0, len(l) - 1)]

text = Rule(re.compile("([^\[\]\|]|\\\[|\\\]|\\\|)+"))

spin = Rule("[", R("toplevel"), ZeroOrMore("|", R("toplevel"), Action(lambda a, b: b)), "]",
    Action(_choose))

toplevel = Rule(ZeroOrMore(Either(text, spin)), Action(lambda l: "".join(l)))

p = Parser(toplevel)

if __name__ == "__main__":
    import string
    import sys
    print(p.parse(" ".join(sys.argv[1:])))





"""
    @author Christophe Eymard <christophe.eymard@ravelsoft.com>
"""

import re

from pwpeg import *
from helpers import *
from pwast import *

re_regexp = re.compile("/(\\/|[^/])+/[idsmlux]+")

def concat(arr):
    if arr is None or len(arr) == 0:
        return ""

    res = []
    for a in arr:
        if isinstance(a, list): 
            res.append(_concat(a))
        else:
            res.append(a)

    return "".join(res)

PIPE = "|"
SPACE = Rule(re.compile("\s*"), name="Whitespaces")
LINE_SPACE = re.compile("[ \t]*")
LPAREN = "("
RPAREN = ")"
LBRACKET = "["
RBRACKET = "]"
LBRACE = "{"
RBRACE = "}"
ARROW = "->"
DOLLAR = "$"
AMPS = "&"
EXCL = "!"
EQUAL = "="
COLON = ":"
COMMA = ","
STAR = "*"
QUESTION = "?"
PLUS = "+"
EOL = "\n"

############################

space_and_comments = re.compile("(\s+|#.*$)*", re.M)

identifier = Rule(re.compile("[a-zA-Z_][a-zA-Z0-9_]*"), name="Identifier")

to_eol = Rule(re.compile("[^\n]*", re.M))

empty_lines = Rule(re.compile("([ \t]*\n)*", re.M))

regexp = Rule(re_regexp)

number = Rule(re.compile("-?[0-9]+"), Action(lambda n: int(n)))

starting_code = Rule("%%", re.compile("((?!%%).)*", re.DOTALL), "%%", Action(lambda b, t, e: t))

##############################################################################################

balanced_paren      = balanced("(", ")")
balanced_braces     = balanced("{", "}")
balanced_brackets   = balanced("[", "]")

# Re-concatenate the string.
string = Either(
    (delimited('\''), Action(lambda x: "".join(x))),
    (delimited('"'), Action(lambda x: "".join(x))),
    # Backslash quoted string
    (re.compile("\\\\[^ \t\n\[\]\|\)]+"), Action(lambda s: "'" + s[1:].replace('\'', '\\\'') + "'"))
)

regexp = Rule(
    delimited('/'), Optional(re.compile('[idsmlux]+')),
    skip=None
)

action = Either(
    Rule(balanced(LBRACE, RBRACE), 
            Action(lambda b: b[1].strip()),
        name="Brace Action"),

    Rule(Optional(SPACE), ARROW, Optional(LINE_SPACE), EOL, indented_block,
            Action(lambda sp, arrow, more_space, eol, code: code),
        skip=None, name="Multi Line Action"),

    Rule(ARROW, to_eol, 
            Action(lambda arrow, line: line.strip()), 
        skip=None, name="Single Line Action")
)

external_rule = Either(
    (DOLLAR, identifier, Action(lambda d, i: i)),
    (DOLLAR, balanced_paren, Action(lambda d, b: concat(b))),
    skip=None
)

#######################################################

predicate = Rule(
    Either(AMPS, EXCL), balanced_braces
)


rulename = Rule(identifier, Optional(balanced_paren), 
    Action(lambda i, b: i + concat(b)), 
    name="Rule Name")

real_rule = Rule(Either(
    regexp,
    string,
    Rule(rulename, Action(lambda x: AstRuleCall(x))),
    external_rule,
    R("either_rule")
), Action(lambda r: AstRuleSingle(r)),
name="Real Rule")

label = Rule(identifier, COLON, Action(lambda name, _2: name), name="Production Label")

repetition = Rule(Either(
    ("*", Action(lambda x: (0, -1))),
    ("+", Action(lambda x: (1, -1))),
    ("?", Action(lambda x: (0,  1))),
    ("<", number, ">", Action(lambda l, n, r: (n, n))),
    ("<", Optional(number), ",", Optional(number), ">", Action(lambda l, fr, c, to, r: (-1 if fr is None else fr, -1 if to is None else to) ))
))

full_rule = Rule(Optional(label), Not(R("ruledecl")), real_rule, Optional(repetition),
        Action(lambda label, rule, rep: rule.set_label(label).set_repetition(rep)),
    name="Full Rule")

match_rule = Rule(Either("!", "&"), real_rule, Optional(repetition),
        Action(lambda sym, rule, mod: rule.set_modifier(mod).set_matching(sym)),
    name="Matching Rule")

either_rule = Rule(
    LBRACKET, OneOrMoreSeparated(R("rules"), PIPE), RBRACKET, Action(lambda _1, rules, _2: AstRuleEither(rules))
)

rule_repeat = Rule(OneOrMore(Either(match_rule, full_rule, predicate)), Optional(action), Action(lambda x, a: AstRuleGroup(x, a)))

rules = Rule(
    OneOrMoreSeparated(rule_repeat, PIPE), Action(lambda x: AstRuleEither(x)),
    name="Rules+"
)

####################### Rule declaration ###########################

rulearg = Rule(identifier, real_rule, 
        Action(lambda id, rule: (id, rule)),
    name="Rule Argument")

ruledecl = Rule(
    rulename, Optional(OneOrMoreSeparated(rulearg, COMMA)), EQUAL, 
    name="Rule Declaration"
)

grammarrule = Rule(
    ruledecl, rules, Action(lambda decl, rules: AstRuleDecl(decl[0], decl[1], rules)),
    name="Grammar Rule"
)

toplevel = Rule(Optional(starting_code), OneOrMore(grammarrule), 
        Action(lambda code, rules: AstFile(code, rules)),
    skip=space_and_comments, name="Top Level")

###################### <<< ACTIONS >>>

@predicate.set_action
def _predicate_action(op, braced_code):
    code = braced_code[1].strip()
    if op == EXCL:
        code = "not({0})".format(code)
    return AstPredicate(code)

@regexp.set_action
def _regexp_action(contents, flags):
    args = []

    args.append('\'' + contents[1].replace('\'', '\\\'') + '\'')

    if flags:
        fs = []
        for f in flags:
            fs.append("re.{0}".format(f.toupper()))
        fs = " & ".join(fs)
        args.append(fs)

    return "re.compile({0})".format(", ".join(args))


#####################################################

parser = Parser(toplevel)

#####################################################

if __name__ == "__main__":
    from optparse import OptionParser
    optparser = OptionParser()

    options, args = optparser.parse_args()

    for a in args:
        f = open(a, "r")
        s = f.read()
        f.close()

        res = parser.parse(s)
        print(res.to_python())




"""
    A simple calculator test.
"""

import re
from pwpeg import *

literal = Rule(re.compile("[0-9]+"), Action(lambda n: int(n)))

parengroup = Either(
    literal,
    ("(", R("addition"), ")", Action(lambda _1, res, _2: res))
)

multiplication = Either(
    (parengroup, Either("/", "*"), parengroup, Action(lambda left, op, right: left / right if op == "/" else left * right)),
    parengroup
)

addition = Either(
    (multiplication, Either("-", "+"), multiplication, Action(lambda left, op, right: left - right if op == "-" else left + right)),
    multiplication,
    skip=re.compile("\s+")
)

#p = Parser(addition, re.compile("\s+"))
p = Parser(addition)

print(p.parse("2 * (2 + 3)"))




import re
from pwpeg import *

class LaterValue(object):
    """
    """

    def __init__(self, default):
        self.default = default

    def set_if_unset(self, value):
        if not self.isset():
            self._value = value

    def value(self):
        if not self.isset():
            return self.default
        return self._value

    def isset(self):
        return "_value" in self.__dict__

@rule
def all_but(but, escape="\\", re_flags=0):
    """
    """

    if not isinstance(but, list):
        but = [but]

    but = [re.escape(a) for a in but]

    if escape:
        escaped = [escape + a for a in but]
        regexp = "({0}|[^{1}])+".format("|".join(escaped), "".join(but))
    else:
        regexp = "[^{0}]+".format("".join(but))

    return re.compile(regexp, re_flags)

@rule(skip=None)
def balanced(start, end, escape="\\"):
    """
    """

    @rule
    def __balanced_inside():
        return Either(
            # Recurse on a balanced expression
            (start, R("balanced_inside", start, end, escape), end, Action(lambda s, m, e: s + m + e)),

            # Or simply gobble up characters that neither start nor end or their
            # backslashed version.
            all_but([start, end])
        )

    return start, ZeroOrMore(__balanced_inside), end, Action(lambda s, l, e: (s, "".join(l), e))

@rule
def delimited(char, escape='\\'):
    """
    """

    return char, all_but(char, escape), char

@rule
def __repeating_separated(rules, separator, at_least, at_most):
    """
    """

    def __action(first, rest):
        return [first] + rest

    if at_most == -1:
        # This is just so that it is equal to -1 in the end.
        at_most = 0

    if at_least == 0:
        return Optional(rules, Repetition(0, at_most - 1, separator, rules, Action(lambda _1, r: r), Action(__action)))

    return rules, Repetition(at_least - 1, at_most - 1, separator, rules, Action(lambda _1, _2: _2)), Action(__action)

@rule
def ZeroOrMoreSeparated(rules, sep):
    return __repeating_separated(rules, sep, 0, -1)

@rule
def OneOrMoreSeparated(rules, sep):
    return __repeating_separated(rules, sep, 1, -1)

@rule
def ExactlySeparated(how_much, rules, sep):
    return __repeating_separated(rules, sep, how_much, how_much)

@rule
def RepetitionSeparated(at_least, at_most, rules, sep):
    return __repeating_separated(rules, sep, at_least, at_most)

re_lead = re.compile("^[ \t]+")
to_eol = Rule(re.compile("[^\n]*", re.M))
empty_lines = Rule(re.compile("([ \t]*\n)*", re.M))

@rule(skip=None)
def indented_block(indentation=1):
    """
    """

    indent = LaterValue(indentation)

    @rule
    def __leading_indent(indent):

        if indent.isset():
            # Since we know how many leading spaces our indentation is at,
            # we want to match its exact number.
            pattern = "[ \t]{{{0}}}".format(indent.value())
            return re.compile(pattern)

        # The first time around, we eat all the spaces.
        # Only after does this rule only eat the leading spaces.
        return re_lead


    @rule
    def __indented_line(indent):

        def action_set_indentation(lead, line, opt):
            ind = len(lead)
     
            indent.set_if_unset(ind)

            # the empty lines are just squizzed out of the parsed text.
            return line

        return __leading_indent(indent), to_eol, Optional(empty_lines), Action(action_set_indentation)


    def _action_concatenate(first_line, others):
        lst = [first_line] + others
        return "\n".join(lst)

    return __indented_line(indent), ZeroOrMore(__indented_line(indent)), Action(_action_concatenate)



"""
    @author Christophe Eymard <christophe@ravelsoft.com>
"""

from re import _pattern_type
import inspect
import sys

if sys.version_info >= (3, 0):
    # Python 3 removed entirely the unicode type, so to have
    # the text isinstance() still work, we just remap it to str (which
    # is unicode anyways)
    unicode = str


class SyntaxError(Exception):
    """ The way the text is parsed is by trial and error.

        When SyntaxError is caught by the parser, it goes up the caller
        chain to get to the first Choice() it finds to keep looking
        for a match.
    """


class IgnoreResult(object):
    """ This class is used by the parsing rules to determine wether to add the
        result of some parsing to the general result.

        Typically, the And and Not rules return this as well as and advancement
        of 0 when they match.
    """


def rule(*args, **kwargs):
    if len(args) == 1:
        return Rule(args[0])
    return lambda x: Rule(x, *args, **kwargs)

class Action(object):
    def __init__(self, fn):
        self.f = fn

    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)


class Results(list):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "{0}:{1}".format(self.name, super(Results, self).__repr__())
        


class Rule(object):
    """ A Grammar rule.
    """

    class ArmedRule(object):
        """ A Grammar rule which rules expression has been evaluated.
            Calling it like a function calls its parse() method with
            the right rules.

            The fact we have to arm rules is to allow the rules to be
            directly parametrizable.
        """

        def __init__(self, rule, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            self.rule = rule

        def __call__(self, text, skip=None):
            return self.rule.parse(text, rules=self.rule.rulefn(*self.args, **self.kwargs), skip=skip)

        def __repr__(self):
            return self.rule.__repr__(repr(self.args))


    def __call__(self, *args, **kwargs):
        """ Shortcut to create an ArmedRule out of the current rule.
        """

        return Rule.ArmedRule(self, *args, **kwargs)


    def __init__(self, *args, **kwargs):

        if len(args) == 0:
            raise Exception("Can not have empty rules")

        if callable(args[0]) and not isinstance(args[0], Rule) and not isinstance(args[0], Rule.ArmedRule):
            # Passing a rule function
            self.rulefn = args[0]
            self.name = args[0].__name__

        elif len(args) == 1 and isinstance(args[0], tuple):
            self.rulefn = lambda: args[0]
            self.name = self.__class__.__name__

        else:
            self.rulefn = lambda: args
            self.name = self.__class__.__name__

        self._process_kwargs(kwargs)

    def _process_kwargs(self, kwargs):
        self.name = kwargs.get("name") or self.name

        if "skip" in kwargs:
            self.set_skip(kwargs.get("skip"))


    def set_skip(self, skip):
        if skip is not None:
            if not isinstance(skip, Rule):
                skip = Rule(skip)()
            else:
                skip = skip()

        self.skip = skip


    def _parse_terminal(self, text, terminal):
        """ Try to match a string in the input.
        """

        if text.startswith(terminal):
            return len(terminal), terminal
        return False


    def _parse_regexp(self, text, reg):
        """ Try to match a regexp and returns its entire match.
        """

        m = reg.match(text)
        if m and len(m.group(0)) > 0:
            return len(m.group(0)), m.group(0)
        return False


    def parse(self, text, rules=None, skip=None):
        """ Execute the rules
        """

        # The rule's skipping rule has precedence over the given skip parameter
        if "skip" in self.__dict__:
            skip = self.skip

        advanced = 0
        results = Results(self.name)

        if not isinstance(rules, tuple):
            rules = (rules,)
        
        for i, r in enumerate(rules):
            subrule_result = None

            if skip:
                try:
                    adv, res = skip(text[advanced:])
                    advanced += adv
                except SyntaxError as e:
                    # Nothing to skip.
                    pass

            # We are checking a simple text terminal
            if isinstance(r, unicode) or isinstance(r, str):
                subrule_result = self._parse_terminal(text[advanced:], r)

            # We have to handle a regexp.
            elif isinstance(r, _pattern_type):
                subrule_result = self._parse_regexp(text[advanced:], r)

            # We have an armed rule, which technically means that the rule function of
            # the rule has already been evaluated.
            elif isinstance(r, Rule.ArmedRule):
                subrule_result = r(text[advanced:], skip=skip)

            # We have a rule that hasn't be armed. We assume that it's rule function doesn't
            # take arguments.
            elif isinstance(r, Rule):
                r2 = r()
                subrule_result = r2(text[advanced:], skip=skip)

                # Text was not matched, so just go to the next rule.
                if subrule_result is None:
                    continue

            elif isinstance(r, Action):
                return advanced, r(*results)

            # Otherwise, this is just a predicate
            else:
                # The predicate is called with the results we have so far.
                # It will raise SyntaxError if it returns false.
                res = r(*results)

                if res is False:
                    raise SyntaxError("The predicate was not satisfied")

                continue

            if subrule_result is False:
                # FIXME We should keep in Either a list of all the tried combination
                # that failed, as well as the position where the failing was
                # found.

                # The subrule_result returning false is only when matching a terminal
                # or a regular expression.
                raise SyntaxError(repr(r))

            # If everything went according to plan, the subrule_result is an tuple
            # with the number of consumed characters and the result of the processing
            # of the rule.
            sub_adv, sub_processed_result = subrule_result

            advanced += sub_adv

            if sub_processed_result is not IgnoreResult:
                results.append(sub_processed_result)

        if "action" in self.__dict__:
            return advanced, self.action(*results)

        if len(results) == 1:
            return advanced, results[0]

        return advanced, results

    
    def set_action(self, fn):
        """ Add an action function to the rules.
            
            This is to be used when the processor function is not directly given to
            the rule in its constructor.
        """

        self.action = fn
        return fn

    def __repr__(self, args=""):
        return "<{0}{1}>".format(self.name, args)



class Repetition(Rule):
    """ A set of rules that can be repeated, like {m,n} in regular expressions.
    """


    def __init__(self, _from, _to, *args, **kwargs):
        self._from = _from
        self._to = _to
        super(Repetition, self).__init__(*args, **kwargs)


    def parse(self, text, rules, skip=None):
        results = []

        times = 0
        _from, _to = self._from, self._to

        advance = 0

        while advance < len(text) and (_to == -1 or times < _to):
            try:
                # Get the resultss.
                adv, res = super(Repetition, self).parse(text[advance:], rules, skip)
            except SyntaxError as e:
                break

            # Parsing was successful, so we add it to the results.
            advance += adv

            if res is not IgnoreResult:
                results.append(res)

            # We repeated one more time !
            times += 1

        if _from != -1 and times < _from:
            raise SyntaxError("Rule needs to be repeated at least {0} times".format(_from))

        return advance, results



class OneOrMore(Repetition):
    """ A Repetition that acts like the + in regular expressions.
    """

    def __init__(self, *args, **kwargs):
        super(OneOrMore, self).__init__(1, -1, *args, **kwargs)



class ZeroOrMore(Repetition):
    """ A Repetition that acts like the * in regular expressions.
    """

    def __init__(self, *args, **kwargs):
        super(ZeroOrMore, self).__init__(0, -1, *args, **kwargs)



class Exactly(Repetition):
    """ A Repetition that wants to match excatly `times` elements.
    """
    
    def __init__(self, times, *args, **kwargs):
        super(Exactly, self).__imit__(times, times, *args, **kwargs)



class Optional(Repetition):
    """ The equivalent of the ? in regular expressions.
    """

    def __init__(self, *args, **kwargs):
        super(Optional, self).__init__(0, 1, *args, **kwargs)


    def parse(self, text, rules, skip=None):
        adv, res = super(Optional, self).parse(text, rules, skip)

        if len(res) == 0:
            return 0, None

        return adv, res[0]



class Not(Rule):
    """ Look ahead in the input. If no syntax error is received, then
        raise a SyntaxError.

        This rule is used to check that a rule does not apply on the input
        following the current position.

        It does *not* advance the parser position.
    """

    def parse(self, text, rules, skip):
        try:
            adv, res = super(Not, self).parse(text, rules, skip)
        except SyntaxError as e:
            # Couldn't match the next rule, which is what we want, so
            # we return a result that won't advance the parser.
            return 0, IgnoreResult
        raise SyntaxError("Matched input that should not have been matched")
                


class And(Rule):
    """ Look ahead in the input. If its rules can be applied, then continue
        parsing from the same place we were before executing this rule.

        This rule is used to check that the following input contains a rule we
        want.

        It does *not* advance the parser position.
    """

    
    def parse(self, text, rules, skip):
        # Try to parse our rules
        super(And, self).parse(text, rules, skip)

        # If there was no error, we don't advance, which is what we want.
        # A syntax error is raised otherwise in our super() parse.
        return 0, IgnoreResult



class Either(Rule):
    """ Try parsing with several rules and return the result of the first
        one that works.

        The rules are given to the constructor as its arguments.
    """

    def __init__(self, *args, **kwargs):
        self.rules = []

        for a in args:
            if isinstance(a, Rule):
                self.rules.append(a)
            else:
                self.rules.append(Rule(a))

        self.rulefn = lambda: None
        self.name = "Either"

        self._process_kwargs(kwargs)


    def parse(self, text, rules, skip):
        if "skip" in self.__dict__:
            skip = self.skip

        for rule in self.rules:
            try:
                r = rule()
                return r(text, skip)
            except SyntaxError as e:
                # FIXME should store the error

                # We continue since the SyntaxError just means that we didn't match and
                # must try the next choice.
                continue

        raise SyntaxError("None of the provided choices matched")



class R(Rule):
    """ A proxy Rule to call rules not yet defined in the scope.

        What it does is by using the `inspect` module, it stores a reference
        to the module where it is instanciated as well as the name of the rule
        in said module. 
        
        When the rule is to be used for parsing, it fetches it from the module
        and executes it.

        This means that this Rule can be used to refer to rules that will be
        defined later in the code, or even the current rule if we need
        recursion.
    """
    
    def __init__(self, rulename, *args, **kwargs):

        c = inspect.currentframe()
        self.gl = c.f_back.f_globals
        self.name = rulename
        self.args = args
        self.kwargs = kwargs


    def parse(self, *args, **kwargs):
        raise Exception("This method should never be called directly.")


    def __call__(self):
        """ Return the result of calling the real rule after fetching
            it from its global object.
        """

        return self.gl[self.name](*self.args, **self.kwargs)

def analyse_frames(f, i=[0]):
    if f.f_back:
        analyse_frames(f.f_back)
    current_obj = f.f_locals.get("self")
    l = f.f_locals
    if isinstance(current_obj, Rule):
        print((" " * i[0]) + repr(current_obj) + " " + repr(l.get("rules")) + " -> " + repr(l.get("results")))
        i[0] += 1

class Parser(object):
    """ A parser that parses a text input.

        It is given a top-level rule with which it will start the parsing,
        as well as a skip rule which will be checked agains before executing
        any rule, useful to remove white spaces and comments.
    """

    def __init__(self, toprule):

        if not isinstance(toprule, Rule):
            toprule = Rule(toprule)

        self.toprule = toprule


    def parse(self, text, *args, **kwargs):
        """ Parse the given input and return the result of the parsing.

            An Exception will be raised if the parsing does not use the
            integrality of the text.
        """

        try:
            parse = self.toprule(*args, **kwargs)
            result = parse(text)
        except Exception as e:
            etype, eobj, etb = sys.exc_info()
            tb = etb
            while tb.tb_next:
                tb = tb.tb_next
            print("The following exception was received:\n")
            import traceback
            traceback.print_tb(etb)
            print(e)

            print("\nWhile parsing, at this point:\n")
            analyse_frames(tb.tb_frame)

            return

        if result[0] != len(text):
            raise Exception("Finished parsing, but all the input was not consumed by the parser. Leftovers: {0}".format(text[result[0]:]))

        # Everything went fine, sending the results.
        return result[1]


    def partial_parse(self, text, *args, **kwargs):
        """ Parse the given input and return only the result of the parsing,
            which is a tuple containing the (number of consumed characters, result of parsing).

            It can be used on some inputs on which we know the grammar can't or won't consume
            the totality of the input.
        """

        parse = self.toprule(*args, **kwargs)
        return parse(text)




import helpers
import pwpeg

class AstNode(object):
    def __repr__(self):
        return "<{0}>".format(self.__class__.__name__)

    def to_python(self, ctx=dict(), indent=0):
        return "-"


class AstPredicate(AstNode):
    def __init__(self, code):
        self.code = code


    def __repr__(self):
        return "{{{0}}}".format(self.code)

    def to_python(self, ctx=dict(), indent=0):
        return "lambda {0}: {1}".format(", ".join(ctx["args"]), self.code)


class AstRuleSingle(AstNode):
    def __init__(self, rule, repetition=None):
        self.label = None
        self.rule = rule
        self.repetition = None
        self.matching = None

    def set_label(self, label):
        self.label = label
        return self

    def set_repetition(self, repetition):
        self.repetition = repetition
        return self

    def set_matching(self, matching):
        self.matching = matching
        return self

    def __repr__(self):
        label = ""
        repetition = self.repetition or ""
        if self.label:
            label = self.label + ":"
        return "{0}{1}{2}".format(label, self.rule, repetition)

    def to_python(self, ctx=dict(), indent=0):
        # FIXME the repetitions !
        res = None

        if isinstance(self.rule, AstNode):
            res = self.rule.to_python(ctx, indent)
        else:
            res = self.rule

        rep = self.repetition
        if rep:
            if rep[0] == rep[1]:
                res = "Exactly({0}, {1})".format(rep[0], res)
            elif rep[0] == 0 and rep[1] == 1:
                res = "Optional({0})".format(res)
            elif (rep[0] == 0 or rep[0] == -1) and rep[1] == -1:
                res = "ZeroOrMore({0})".format(res)
            elif rep[0] == 1 and rep[1] == -1:
                res = "OneOrMore({0})".format(res)
            else:
                res = "Repetition({0}, {1}, {2})".format(rep[0], rep[1], res)

        if self.matching == "!":
            res = "Not({0})".format(res)
        elif self.matching == "&":
            res = "And({0})".format(res)
        else:
            res = res

        return res


class AstRuleCall(AstRuleSingle):
    def to_python(self, ctx, indent):
        rulename = self.rule

        paren_start = self.rule.find("(")
        
        if paren_start != -1:
            rulename = self.rule[:paren_start]

        if not rulename in ctx["seen_rules"]:
            if paren_start != -1:
                self.rule = "R(\'{0}\', {1}".format(rulename, self.rule[paren_start + 1:])
            else:
                self.rule = "R(\'{0}\')".format(rulename)

        return super(AstRuleCall, self).to_python(ctx, indent)


class AstCode(AstNode):
    def __init__(self, code):
        self.code = code

    def __repr__(self, ctx=dict(), indent=0):
        code = ""
        if self.code:
            code = self.code[:10]
            if len(self.code) > 10:
                code += "..."
        return code
        


class AstRuleGroup(AstNode):
    def __init__(self, rules, action=None):
        self.rules = rules
        self.action = action

    def set_action(self, action):
        self.action = action
        return self

    def __repr__(self):
        rules = " ".join([repr(r) for r in self.rules])

        action = ""

        if self.action:
            action = " ({0})".format(repr(self.action))

        return "{0}{1}".format(rules, action)

    def to_python(self, ctx=dict(), indent=0):
        ctx = dict(**ctx)
        ctx["args"] = []

        res = []

        i = 0
        for r in self.rules:

            if not getattr(r, "matching", None) and not isinstance(r, AstPredicate):
                ctx["args"].append(getattr(r, "label", None) or "_{0}".format(i))
                i += 1
            
            res.append(r.to_python(ctx, indent))

        if self.action:
            if self.action.find('\n') is -1 and self.action.find('return') is -1:
                res.append("Action(lambda {0}: {1})".format(", ".join(ctx["args"]), self.action))
            else:
                action_name = "action_{0}".format(ctx["last_action"][0])
                res.append("Action({0})".format(action_name))
                ctx["last_action"][0] += 1

                action_body = []

                action_body.append("def {0}({1}):".format(action_name, ", ".join(ctx["args"])))
                for l in self.action.split("\n"):
                    action_body.append("    " + l)
                action_body.append("")

                ctx["actions"].append("\n".join(action_body))

        if len(res) > 1:
            return "(" + ", ".join(res) + ")"
        return res[0]



class AstRuleEither(AstRuleGroup):
    def __repr__(self):
        rep = " | ".join([repr(r) for r in self.rules])
        if len(self.rules) > 1:
            return "[" + rep + "]"
        return rep

    def to_python(self, ctx=dict(), indent=0):
        if len(self.rules) == 1:
            return self.rules[0].to_python(ctx, indent)

        normal_ind = " " * indent
        ind = " " * (indent + 4)

        return "Either(\n{0}{1}\n{2})".format(ind, 
            ",\n{0}".format(ind).join([r.to_python(ctx, indent + 4) for r in self.rules]),
            normal_ind)


class AstRuleDecl(AstNode):
    def __init__(self, name, rules, skip=None):
        self.name = name
        self.rules = rules
        self.skip = skip

    def __repr__(self):
        args = ""
        if self.args:
            args = ", ".join([o[0] for o in self.args])
            args = "({0})".format(args)
        return "<{0}{1}> {2}".format(self.name, args, self.rules)

    def to_python(self, ctx=dict(), indent=0):
        res = []

        ctx = dict(**ctx)
        actions = ctx["actions"] = []

        skip = ("skip=" + self.skip.to_python(ctx, indent) if self.skip else "")

        # The rule has some parameters.
        if self.name.endswith(")"):
            res.append("@rule({0})".format(skip))
            res.append("def {0}:".format(self.name))

            result = "    return {1}".format(self.name, self.rules.to_python(ctx, indent + 4))

            if actions:
                actions = ["\n".join(["    " + l for l in a.split('\n')]) for a in actions]
                res.append("\n\n".join(actions))

            res.append(result)
        else:
            if skip: skip = ", {0}".format(skip)
            res.append("{0} = Rule({1}, name='{0}'{2})".format(self.name, self.rules.to_python(ctx, indent), skip))

            if actions:
                res.insert(0, "\n\n".join(actions))

        ctx["seen_rules"].append(self.name)
        return "\n".join(res)
            


class AstFile(AstNode):

    def __init__(self, code, rules):
        self.code = code or ""
        self.rules = rules

    def __repr__(self):
        return "{0}{1}".format(self.code, self.rules)

    def to_python(self, ctx=dict(), indent=0):
        ctx["last_action"] = [0]
        ctx["seen_rules"] = [n for n in helpers.__dict__.keys()] + \
            [n for n in pwpeg.__dict__.keys()] + ["None"]

        return "\n".join(["import re\n",
            "from pwpeg import *",
            "from pwpeg.helpers import *",
            "\n" + self.code,
            "\n\n".join([r.to_python(ctx, indent) for r in self.rules])
        ])

